from typing import List
from typing import cast

from dollar.dollarexception import DollarExecutionException
from dollar.dollarexception import DollarException
from dollar.dollarobject import DollarObject
from dollar.helper.validationhelper import ValidationHelper
from dollar.format.input.inputformat import InputFormat
from dollar.format.input.inputformat import InputFormatText
from dollar.format.input.inputformat import InputFormatDollarObject
from dollar.format.input.inputformat import InputFormatDollarObjectValue
from dollar.format.input.inputformat import InputFormatFunction
from dollar.format.input.inputformat import InputFormatBlock
from dollar.format.input.inputformat import InputFormatUnion
from dollar.format.input.inputformattype import InputFormatType
from dollar.format.output.outputfactory import OutputFactory
from dollar.format.output.outputformatterhandler import OutputFormatterHandler
from dollar.plugin.pluginargvalidatior import PluginArgValidator


class InputToStrTransformer:

    @classmethod
    def transform(cls, this_dollar_object: DollarObject, input_format: InputFormat) -> str:
        if input_format.get_format_type() == InputFormatType.TEXT:
            input_format = cast(InputFormatText, input_format)
            return input_format.get_text()

        elif input_format.get_format_type() == InputFormatType.DOLLAR_OBJECT:
            input_format = cast(InputFormatDollarObject, input_format)
            return OutputFactory.create_link_dollar_object(this_dollar_object, input_format.get_dollar_object())

        elif input_format.get_format_type() == InputFormatType.DOLLAR_OBJECT_VALUE:
            input_format = cast(InputFormatDollarObjectValue, input_format)
            if type(input_format.get_value()) == str:
                if not ValidationHelper.valid_str(input_format.get_value()):
                    raise DollarExecutionException(
                            "When rendering, dollar header referenced value needs to be a non blank str",
                            input_format.get_dollar_context())
                return input_format.get_value()
            elif isinstance(input_format.get_value(), DollarObject):
                return OutputFactory.create_link_dollar_object(this_dollar_object, input_format.get_value())
            else:
                raise DollarExecutionException(
                        "When rendering, dollar header referenced value needs to be str or DollarObject",
                        input_format.get_dollar_context())

        elif input_format.get_format_type() == InputFormatType.FUNCTION:
            input_format = cast(InputFormatFunction, input_format)
            plugin = input_format.get_function_plugin()
            params = input_format.get_parameters()
            plugin.get_arg_info()

            try:
                PluginArgValidator.validate(params, plugin.get_arg_info())
                plugin_output = plugin.exec_function(*params)
                return OutputFormatterHandler.get_formatter().format(this_dollar_object, plugin_output)
            except DollarException as e:
                raise DollarExecutionException(
                        e.get_message(),
                        input_format.get_dollar_context()) from e

        elif input_format.get_format_type() == InputFormatType.BLOCK:
            input_format = cast(InputFormatBlock, input_format)
            plugin = input_format.get_block_plugin()
            content = input_format.get_content()
            plugin_output = plugin.exec_block(content)
            try:
                return OutputFormatterHandler.get_formatter().format(this_dollar_object, plugin_output)
            except DollarException as e:
                raise DollarExecutionException(
                        e.get_message(),
                        input_format.get_dollar_context()) from e

        elif input_format.get_format_type() == InputFormatType.UNION:
            input_format = cast(InputFormatUnion, input_format)
            return cls.transform_list(this_dollar_object, input_format.get_children())

        else:
            raise DollarExecutionException(
                    "Format {} is not supported".format(input_format.get_format_type()),
                    input_format.get_dollar_context())

    @classmethod
    def transform_list(cls, this_dollar_object: DollarObject, input_formats: List[InputFormat]):
        result = []
        for input_format in input_formats:
            result.append(cls.transform(this_dollar_object, input_format))
        return "".join(result)



from .linkfunctionplugin import LinkFunctionPlugin
from .listbacklinksfunctionplugin import ListBackLinksFunctionPlugin
from .listfunctionplugin import ListFunctionPlugin
from .listreffunctionplugin import ListRefFunctionPlugin
from .pageextensionplugin import PageExtensionPlugin



from dollar.dollarcontext import DollarContext


class DollarException(Exception):

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def get_message(self):
        return self.message


class DollarExecutionException(DollarException):

    def __init__(self, message: str, dollar_context: DollarContext):
        super().__init__(
                "\n" + dollar_context.get_file() + " on line " + str(dollar_context.get_line()) + ": " + message)
        self.dollar_context = dollar_context

    def get_dollar_context(self):
        return self.dollar_context



class HeaderParserResult:

    def __init__(self, header_object, header_end):
        self.header_object = header_object
        self.header_end = header_end

    def get_header_object(self):
        return self.header_object

    def get_header_end(self):
        return self.header_end



class DollarContext:

    def __init__(self, file: str, line: int, col: int):
        self.file = file
        self.line = line
        self.col = col

    def get_file(self) -> str:
        return self.file

    def get_line(self) -> int:
        return self.line

    def get_col(self) -> int:
        return self.col

    def __str__(self):
        return "DollarContext(" + ", ".join([self.file, str(self.line), str(self.col)]) + ")"


class DollarContextSpan(DollarContext):

    def __init__(self, context_start, context_end):
        super().__init__(context_start.file, context_start.line, context_start.col)
        self.context_start = context_start
        self.context_end = context_end

    def get_context_start(self) -> DollarContext:
        return self.context_start

    def get_context_end(self) -> DollarContext:
        return self.context_end

    def get_file(self) -> str:
        return self.get_context_start().get_file()

    def get_line(self) -> int:
        return self.get_context_start().get_line()

    def get_col(self) -> int:
        return self.get_context_start().get_col()

    def __str__(self):
        return "DollarContextSpan(" \
               + "\n        " + str(self.context_start) \
               + "\n        " + str(self.context_end) + ")"



from dollar.plugin.pluginhandler import PluginHandler
from dollar.plugin.builtin.linkfunctionplugin import LinkFunctionPlugin
from dollar.plugin.builtin.listbacklinksfunctionplugin import ListBackLinksFunctionPlugin
from dollar.plugin.builtin.listfunctionplugin import ListFunctionPlugin
from dollar.plugin.builtin.listreffunctionplugin import ListRefFunctionPlugin
from dollar.plugin.builtin.pageextensionplugin import PageExtensionPlugin


class BuiltinPluginLoader:

    @staticmethod
    def load():
        PluginHandler.register(LinkFunctionPlugin())
        PluginHandler.register(ListBackLinksFunctionPlugin())
        PluginHandler.register(ListFunctionPlugin())
        PluginHandler.register(ListRefFunctionPlugin())
        PluginHandler.register(PageExtensionPlugin())



from dollar.format.output.outputformatter import OutputFormatterMarkdown


class OutputFormatterHandler:
    formatter = OutputFormatterMarkdown()

    @classmethod
    def get_formatter(cls):
        return cls.formatter



from .outputfactory import OutputFactory
from .outputformat import OutputFormat
from .outputformat import OutputFormatBlock
from .outputformat import OutputFormatText
from .outputformat import OutputFormatBold
from .outputformat import OutputFormatItalic
from .outputformat import OutputFormatCode
from .outputformat import OutputFormatLink
from .outputformat import OutputFormatHeader
from .outputformat import OutputFormatParagraph
from .outputformat import OutputFormatOrderedList
from .outputformat import OutputFormatUnorderedList
from .outputformat import OutputFormatListItem
from .outputformat import OutputFormatCodeBlock
from .outputformat import OutputFormatQuoteBlock
from .outputformat import OutputFormatDefinition
from .outputformat import OutputFormatPluginBlock
from .outputformat import OutputFormatDollarObject
from .outputformatter import OutputFormatterMarkdown
from .outputformatterhandler import OutputFormatterHandler
from .outputformattype import OutputFormatType



import sys

from dollar.builder.dollarbuilder import DollarBuilder


def cli():
    args = sys.argv[1:]
    if len(args) == 0:
        print("You need to provide atleast one option")
        print()
        args = ["help"]
    main_option = args[0]
    if main_option == "help":
        print("dollardoc help")
        print("dollardoc build")
        print()
    elif main_option == "build":
        DollarBuilder.build()



from dollar.dollarexception import DollarException
from dollar.dollarobject import DollarObject
from dollar.helper.validationhelper import ValidationHelper


class DollarObjectIdMap:

    id_map = {}

    @classmethod
    def get(cls, dollar_object_id):
        if not ValidationHelper.valid_str(dollar_object_id):
            raise DollarException(
                    "Id must be a valid string")
        if dollar_object_id not in cls.id_map:
            raise DollarException("Id " + dollar_object_id + " does not exist")
        return cls.id_map[dollar_object_id]

    @classmethod
    def add(cls, dollar_object: DollarObject):
        if not ValidationHelper.valid_obj(dollar_object, DollarObject):
            raise DollarException("Object is not a valid DollarObject")
        if not ValidationHelper.valid_str(dollar_object.get_id()):
            raise DollarException("Id must be a valid string")
        if dollar_object.get_id() in cls.id_map:
            raise DollarException("Id " + dollar_object.get_id() + " already exists")
        cls.id_map[dollar_object.get_id()] = dollar_object

    @classmethod
    def clean(cls):
        cls.id_map = {}



from .confighandler import ConfigHandler
from .dollarcontext import DollarContext
from .dollarexception import DollarExecutionException
from .dollarobject import DollarObject
from .dollarobjectidmap import DollarObjectIdMap



from dollar.dollarexception import DollarExecutionException
from dollar.dollarcontext import DollarContext
from dollar.helper.validationhelper import ValidationHelper


class DollarFile:

    def __init__(self, path, content):
        if not ValidationHelper.valid_str(path):
            raise DollarExecutionException(
                    "Path passed to DollarFile must be valid string",
                    DollarContext(path, 0, 0))
        if not ValidationHelper.valid_str(content):
            raise DollarExecutionException(
                    "Content passed to DollarFile must be valid string",
                    DollarContext(path, 0, 0))
        self.path = path
        self.content = content

    def get_path(self):
        return self.path

    def get_content(self):
        return self.content



import yaml
from enum import Enum

from dollar.dollarexception import DollarExecutionException


class ConfigType(str, Enum):
    DOCS_PATH = "docs_path"  # str
    TARGET_PATH = "target_path"  # str
    PLUGIN_PATH = "plugin_path"  # str
    FILE_PASSTHROUGH = "file_passthrough"  # List[str]


class ConfigHandler:
    config = {}
    required_config = [
        ConfigType.DOCS_PATH,
    ]

    @classmethod
    def set_config(cls, config):
        cls.config = config
        cls.validate()

    @classmethod
    def get(cls, key):
        if type(key) == str:
            raise DollarExecutionException("Keys must be provided in type ConfigType")
        if key not in cls.config:
            raise DollarExecutionException("Key {} was not provided in config".format(key))
        return cls.config[key]

    @classmethod
    def get_str_list(cls, key):
        value = cls.get(key)
        if type(value) == str:
            return [value]
        elif type(value) == list:
            return value
        else:
            raise DollarExecutionException("Key {} was not list or string as expected".format(key))

    @classmethod
    def get_opt(cls, key, default):
        if type(key) == str:
            raise DollarExecutionException("Keys must be provided in type ConfigType")
        if key in cls.config:
            return cls.config[key]
        return default

    @classmethod
    def get_str_list_opt(cls, key):
        value = cls.get_opt(key, [])
        if type(value) == str:
            return [value]
        elif type(value) == list:
            return value
        else:
            raise DollarExecutionException("Key {} was not list or string as expected".format(key))

    @classmethod
    def load_config_default(cls):
        cls.load_config_file("dollarconfig.yaml")

    @classmethod
    def load_config_file(cls, path):
        with open(path, "r") as f:
            cls.config = yaml.safe_load(f.read())
        cls.validate()

    @classmethod
    def get_plugin_config(cls, plugin_name):
        if "plugin" not in cls.config:
            return None
        if plugin_name not in cls.config["plugin"]:
            return None
        return cls.config["plugin"][plugin_name]

    @classmethod
    def validate(cls):
        for req in cls.required_config:
            if req not in cls.config:
                raise DollarExecutionException("Required config {} not present in config".format(req))



from .dollarplugin import DollarPlugin
from .dollarplugin import DollarFunctionPlugin
from .dollarplugin import DollarBlockPlugin
from .dollarplugin import DollarExtensionPlugin
from .pluginarg import PluginArg
from .pluginarg import PluginArgDollarObject
from .pluginhandler import PluginHandler
from .plugintype import PluginType



from typing import List
from typing import cast

from dollar.dollarexception import DollarExecutionException
from dollar.dollarexception import DollarException
from dollar.dollarcontext import DollarContext
from dollar.dollarobject import DollarObject
from dollar.dollarobjectidmap import DollarObjectIdMap
from dollar.format.input.inputformat import InputFormatText
from dollar.format.input.inputformat import InputFormatDollarObject
from dollar.format.input.inputformat import InputFormatDollarObjectValue
from dollar.format.input.inputformat import InputFormatFunction
from dollar.format.input.inputformat import InputFormatBlock
from dollar.format.input.inputformat import InputFormatUnion
from dollar.format.input.inputformattype import InputFormatType
from dollar.format.raw.rawdollarformat import RawDollarFormat
from dollar.format.raw.rawdollarformat import RawDollarFormatText
from dollar.format.raw.rawdollarformat import RawDollarFormatReference
from dollar.format.raw.rawdollarformat import RawDollarFormatFunction
from dollar.format.raw.rawdollarformat import RawDollarFormatBlock
from dollar.format.raw.rawdollarformat import RawDollarFormatUnion
from dollar.format.raw.rawdollarformattype import RawDollarFormatType
from dollar.helper.dollarobjecthelper import DollarObjectHelper
from dollar.plugin.pluginmap import PluginMap


class _DollarReferenceObject:

    def __init__(self, dollar_object: DollarObject, value=None):
        self.dollar_object = dollar_object
        self.value = value

    def getdollarobject(self) -> DollarObject:
        return self.dollar_object

    def getvalue(self):
        return self.value

    def hasvalue(self):
        return self.value is not None


class RawToInputTransformer:

    @classmethod
    def transform(cls, this_dollar_object: DollarObject, raw_format: RawDollarFormat):
        if raw_format.get_format_type() == RawDollarFormatType.TEXT:
            raw_format = cast(RawDollarFormatText, raw_format)
            return InputFormatText(raw_format.get_text(), raw_format.get_dollar_context())

        elif raw_format.get_format_type() == RawDollarFormatType.REFERENCE:
            raw_format = cast(RawDollarFormatReference, raw_format)
            dollar_reference_object = cls._handle_dollar_parse(
                    this_dollar_object,
                    raw_format.get_target_text(),
                    raw_format.get_dollar_context())
            if dollar_reference_object.hasvalue():
                return InputFormatDollarObjectValue(
                        dollar_reference_object.getdollarobject(),
                        dollar_reference_object.getvalue(),
                        raw_format.get_dollar_context())
            else:
                dollar_reference_object.getdollarobject().add_backlink(this_dollar_object)
                return InputFormatDollarObject(
                        dollar_reference_object.getdollarobject(),
                        raw_format.get_dollar_context())

        elif raw_format.get_format_type() == RawDollarFormatType.FUNCTION:
            raw_format = cast(RawDollarFormatFunction, raw_format)
            try:
                plugin = PluginMap.get_function(raw_format.get_function_name())
            except DollarException as e:
                raise DollarExecutionException(e.get_message(), raw_format.get_dollar_context())
            transformed_params = cls.transform_list(this_dollar_object, raw_format.get_parameters())
            params = []
            for transformed_param in transformed_params:
                if transformed_param.get_format_type() == InputFormatType.TEXT:
                    transformed_param = cast(InputFormatText, transformed_param)
                    params.append(transformed_param.get_text().strip())
                elif transformed_param.get_format_type() == InputFormatType.DOLLAR_OBJECT:
                    transformed_param = cast(InputFormatDollarObject, transformed_param)
                    params.append(transformed_param.get_dollar_object())
                elif transformed_param.get_format_type() == InputFormatType.DOLLAR_OBJECT_VALUE:
                    transformed_param = cast(InputFormatDollarObjectValue, transformed_param)
                    params.append(transformed_param.get_value())
                elif transformed_param.get_format_type() == InputFormatType.UNION:
                    transformed_param = cast(InputFormatUnion, transformed_param)
                    param_result = ""
                    for child in transformed_param.get_children():
                        if child.get_format_type() == InputFormatType.TEXT:
                            child = cast(InputFormatText, child)
                            param_result = param_result + child.get_text()
                        elif child.get_format_type() == InputFormatType.DOLLAR_OBJECT:
                            child = cast(InputFormatDollarObject, child)
                            child_dollar_object = child.get_dollar_object()
                            child_title = DollarObjectHelper.get_title(child_dollar_object)
                            param_result = param_result + child_title
                        elif child.get_format_type() == InputFormatType.DOLLAR_OBJECT_VALUE:
                            child = cast(InputFormatDollarObjectValue, child)
                            if type(child.get_value()) == str:
                                param_result = param_result + child.get_value()
                            else:
                                raise DollarExecutionException(
                                        "Type of value needs to be str when rendering Union parameter value, was {}"
                                        .format(type(child.get_value())),
                                        child.get_dollar_context())
                        else:
                            raise DollarExecutionException(
                                    "Format type {} is not supported when rendering Union parameter"
                                    .format(child.get_format_type()),
                                    child.get_dollar_context())
                    params.append(param_result.strip())
                else:
                    raise DollarExecutionException(
                            "Format type {} is not supported when rendering parameter"
                            .format(transformed_param.get_format_type()),
                            transformed_param.get_dollar_context())
            return InputFormatFunction(plugin, params, this_dollar_object, raw_format.get_dollar_context())

        elif raw_format.get_format_type() == RawDollarFormatType.BLOCK:
            raw_format = cast(RawDollarFormatBlock, raw_format)
            try:
                plugin = PluginMap.get_block(raw_format.get_block_name())
            except DollarException as e:
                raise DollarExecutionException(e.get_message(), raw_format.get_dollar_context())
            content = cls.transform(this_dollar_object, raw_format.get_content())
            return InputFormatBlock(plugin, content, raw_format.get_dollar_context())

        elif raw_format.get_format_type() == RawDollarFormatType.UNION:
            raw_format = cast(RawDollarFormatUnion, raw_format)
            children = cls.transform_list(this_dollar_object, raw_format.get_children())
            return InputFormatUnion(children, raw_format.get_dollar_context())

        else:
            raise DollarExecutionException(
                    "Format {} is not supported".format(raw_format.get_format_type()),
                    raw_format.get_dollar_context())

    @classmethod
    def transform_list(cls, this_dollar_object: DollarObject, raw_formats: List[RawDollarFormat]):
        result = []
        for raw_format in raw_formats:
            result.append(cls.transform(this_dollar_object, raw_format))
        return result

    @classmethod
    def _handle_dollar_parse(
            cls,
            this_dollar_object: DollarObject,
            dollar_str: str,
            dollar_context: DollarContext) -> _DollarReferenceObject:
        dollar_str_split = dollar_str.split(".")
        dollar_id = dollar_str_split[0]
        if dollar_id == "this":
            dollar_object = this_dollar_object
        else:
            try:
                dollar_object = DollarObjectIdMap.get(dollar_id)
            except DollarException as e:
                raise DollarExecutionException(e.get_message(), dollar_context) from e
        check = dollar_object.get_header()
        value = None
        for i in range(1, len(dollar_str_split)):
            key = dollar_str_split[i]
            if key not in check:
                raise DollarExecutionException(
                        "Key {} could not be found in {}"
                        .format(".".join(dollar_str_split[1:]), dollar_id),
                        dollar_context)
            value = check[key]
            check = value
        return _DollarReferenceObject(dollar_object, value)



from dollar.plugin.pluginmap import PluginMap
from dollar.dollarobject import DollarObject
from dollar.helper.validationhelper import ValidationHelper


class DollarObjectHelper:

    @staticmethod
    def is_type(dollar_object: DollarObject, dollar_object_type: str):
        if dollar_object.get_type() == dollar_object_type:
            return True
        dollar_object_type_temp = dollar_object.get_type()
        while PluginMap.has_extension(dollar_object_type_temp):
            dollar_object_type_extends = PluginMap.get_extension(dollar_object_type_temp).extends()
            if dollar_object_type == dollar_object_type_extends:
                return True
            dollar_object_type_temp = dollar_object_type_extends
        return False

    @staticmethod
    def get_title(dollar_object: DollarObject):
        header = dollar_object.get_header()
        if "title" in header:
            title = header["title"]
            if ValidationHelper.valid_str(title):
                return title
        return dollar_object.get_id()



from dollar.dollarobject import DollarObject


class DollarObjectImpl(DollarObject):

    def __init__(self, content, path, output_path, target_path):
        super().__init__(path, output_path, target_path)

        self.content = content.split("\n")
        self.header_end = 0
        self.header = None
        self.unparsed_header = {}
        self.raw_formats = []
        self.input_formats = []
        self.output = ""

    def get_id(self):
        if self.header is None:
            return self.unparsed_header["id"]
        return self.header["id"]

    def get_type(self):
        if self.header is None:
            return self.unparsed_header["type"]
        return self.header["type"]

    def get_header(self):
        return self.header

    def set_header(self, header):
        self.header = header

    def get_header_end(self):
        return self.header_end

    def set_header_end(self, header_end):
        self.header_end = header_end

    def get_unparsed_header(self):
        return self.unparsed_header

    def set_unparsed_header(self, unparsed_header):
        self.unparsed_header = unparsed_header

    def get_content_without_header(self):
        return "\n".join(self.content[self.header_end + 1:])

    def get_output(self) -> str:
        return self.output

    def set_output(self, output: str):
        self.output = output

    def get_path(self):
        return self.path

    def get_output_path(self):
        return self.output_path

    def get_target_path(self):
        return self.target_path

    def get_raw_formats(self):
        return self.raw_formats

    def set_raw_formats(self, raw_formats):
        self.raw_formats = raw_formats

    def get_input_formats(self):
        return self.input_formats

    def set_input_formats(self, input_formats):
        self.input_formats = input_formats



import os
import importlib.util

from dollar.plugin.dollarplugin import DollarPlugin
from dollar.plugin.pluginmap import PluginMap


class PluginHandler:

    @classmethod
    def _import_plugin_path(cls, path):
        spec = importlib.util.spec_from_file_location(path, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

    @classmethod
    def import_plugins(cls, path):
        cls._import_plugins(os.path.join(os.path.curdir, path))

    @classmethod
    def _import_plugins(cls, abs_path):
        dir_list = os.listdir(abs_path)
        for item in dir_list:
            new_path = os.path.join(abs_path, item)
            if os.path.isdir(new_path):
                cls.import_plugins(new_path)
            elif item.endswith(".py"):
                cls._import_plugin_path(new_path)

    @classmethod
    def register(cls, plugin: DollarPlugin):
        PluginMap.add(plugin)

    @classmethod
    def clean(cls):
        PluginMap.clean()



import os

from dollar.dollarexception import DollarException
from dollar.file.dollarfile import DollarFile


class DollarFileReader:

    @classmethod
    def read_mdd_files(cls, src):
        return cls._read_mdd_files(src)

    @classmethod
    def _read_mdd_files(cls, open_dir, _open_dir_start=None):
        if _open_dir_start is None:
            _open_dir_start = open_dir
        result = []
        dir_list = os.listdir(open_dir)
        for item in dir_list:
            full_path = os.path.join(open_dir, item)
            if os.path.isdir(full_path):
                result2 = cls._read_mdd_files(full_path, _open_dir_start)
                result = result + result2
            elif item.endswith(".mdd"):
                try:
                    f = open(full_path, "r")
                except Exception as e:
                    raise DollarException("Failed to open file {}".format(full_path)) from e
                full_path_split = os.path.normpath(full_path).split(os.path.sep)
                open_dir_start_split = os.path.normpath(_open_dir_start).split(os.path.sep)
                to = len(full_path_split)
                if to > len(open_dir_start_split):
                    to = len(open_dir_start_split)
                new_path = str(os.path.sep).join(full_path_split[to:])
                try:
                    file_output = f.read()
                except Exception as e:
                    raise DollarException("Failed to read file {}".format(full_path)) from e
                result.append(DollarFile(new_path, file_output))
                f.close()
            else:
                pass  # Allow other files to exist in src
        return result



from typing import List

from dollar.dollarexception import DollarExecutionException
from dollar.dollarcontext import DollarContext
from dollar.helper.validationhelper import ValidationHelper
from dollar.format.raw.rawdollarformattype import RawDollarFormatType


class RawDollarFormat:

    def __init__(self, format_type: RawDollarFormatType, dollar_context: DollarContext):
        self.format_type = format_type
        self.dollar_context = dollar_context

    def get_format_type(self) -> RawDollarFormatType:
        return self.format_type

    def validate(self):
        pass

    def get_dollar_context(self) -> DollarContext:
        return self.dollar_context

    def __repr__(self):
        return str(self.__dict__)


class RawDollarFormatText(RawDollarFormat):

    def __init__(self, text: str, dollar_context: DollarContext):
        super().__init__(RawDollarFormatType.TEXT, dollar_context)
        self.text = text

    def get_text(self) -> str:
        return self.text

    def validate(self):
        if ValidationHelper.valid_str(self.text):
            raise DollarExecutionException("Text can not be blank")


class RawDollarFormatReference(RawDollarFormat):

    def __init__(self, target_text: str, dollar_context: DollarContext):
        super().__init__(RawDollarFormatType.REFERENCE, dollar_context)
        self.target_text = target_text

    def get_target_text(self) -> str:
        return self.target_text

    def validate(self):
        if ValidationHelper.valid_str(self.target_text):
            raise DollarExecutionException("Target text can not be blank")


class RawDollarFormatFunction(RawDollarFormat):

    def __init__(self, function_name: str, parameters: List[RawDollarFormat], dollar_context: DollarContext):
        super().__init__(RawDollarFormatType.FUNCTION, dollar_context)
        self.function_name = function_name
        self.parameters = parameters

    def get_function_name(self) -> str:
        return self.function_name

    def get_parameters(self) -> List[RawDollarFormat]:
        return self.parameters

    def validate(self):
        if ValidationHelper.valid_str(self.function_name):
            raise DollarExecutionException("Function name can not be blank")
        for param in self.parameters:
            param.validate()


class RawDollarFormatBlock(RawDollarFormat):

    def __init__(self, block_name: str, content: RawDollarFormat, dollar_context: DollarContext):
        super().__init__(RawDollarFormatType.BLOCK, dollar_context)
        self.block_name = block_name
        self.content = content

    def get_block_name(self) -> str:
        return self.block_name

    def get_content(self) -> RawDollarFormat:
        return self.content

    def validate(self):
        if ValidationHelper.valid_str(self.block_name):
            raise DollarExecutionException("Block name can not be blank")
        self.content.validate()


class RawDollarFormatUnion(RawDollarFormat):

    def __init__(self, children: List[RawDollarFormat], dollar_context: DollarContext):
        super().__init__(RawDollarFormatType.UNION, dollar_context)
        self.children = children

    def get_children(self) -> List[RawDollarFormat]:
        return self.children

    def validate(self):
        for child in self.children:
            if child.get_format_type() == RawDollarFormatType.UNION:
                raise DollarExecutionException("A union cant contain another union")
            child.validate()



from typing import cast

from dollar.dollarexception import DollarException
from dollar.plugin.dollarplugin import DollarPlugin
from dollar.plugin.dollarplugin import DollarFunctionPlugin
from dollar.plugin.dollarplugin import DollarBlockPlugin
from dollar.plugin.dollarplugin import DollarExtensionPlugin
from dollar.plugin.plugintype import PluginType


class PluginMap:

    function_map = {}
    block_map = {}
    extension_map = {}
    extension_secondary_map = {}

    @classmethod
    def get_function(cls, plugin_name: str) -> DollarFunctionPlugin:
        if plugin_name not in cls.function_map:
            raise DollarException("Function plugin with name {}, cannot be found".format(plugin_name))
        return cls.function_map[plugin_name]

    @classmethod
    def get_block(cls, plugin_name: str) -> DollarBlockPlugin:
        if plugin_name not in cls.block_map:
            raise DollarException("Block plugin with name {}, cannot be found".format(plugin_name))
        return cls.block_map[plugin_name]

    @classmethod
    def get_extension(cls, plugin_name) -> DollarExtensionPlugin:
        if not cls.has_extension(plugin_name):
            raise DollarException("Extension plugin with name {}, cannot be found".format(plugin_name))
        return cls.extension_map[plugin_name]

    @classmethod
    def has_extension(cls, plugin_name):
        return plugin_name in cls.extension_map

    @classmethod
    def get_extension_from_secondary_key(cls, key) -> DollarExtensionPlugin:
        if key not in cls.extension_secondary_map:
            raise DollarException("No extension plugin with handler for secondary key {}".format(key))
        return cls.extension_secondary_map[key]

    @classmethod
    def has_extension_with_secondary_key(cls, key) -> bool:
        return key in cls.extension_secondary_map

    @classmethod
    def add(cls, plugin: DollarPlugin):
        if plugin.get_type() == PluginType.FUNCTION:
            plugin = cast(DollarFunctionPlugin, plugin)
            cls.function_map[plugin.get_name()] = plugin
        elif plugin.get_type() == PluginType.BLOCK:
            plugin = cast(DollarBlockPlugin, plugin)
            cls.block_map[plugin.get_name()] = plugin
        elif plugin.get_type() == PluginType.EXTENSION:
            plugin = cast(DollarExtensionPlugin, plugin)
            cls.extension_map[plugin.get_name()] = plugin
            for secondary in plugin.get_secondaries():
                cls.extension_secondary_map[secondary] = plugin
        else:
            raise DollarException("Plugin with type {} is not supported".format(plugin.get_type()))

    @classmethod
    def clean(cls):
        cls.function_map = {}
        cls.block_map = {}
        cls.extension_map = {}
        cls.extension_secondary_map = {}




import copy

from dollar.dollarexception import DollarExecutionException
from dollar.dollarexception import DollarException
from dollar.dollarcontext import DollarContext
from dollar.dollarobject import DollarObject
from dollar.dollarobjectidmap import DollarObjectIdMap


class HeaderTransformer:

    @classmethod
    def transform(cls, this_dollar_object: DollarObject, header, dollar_context: DollarContext):
        header = copy.deepcopy(header)
        loops = header
        if type(header) == list:
            loops = range(0, len(header))
        for key in loops:
            if key == "id" or key == "type":
                continue

            if type(header[key]) == dict:
                header[key] = cls.transform(this_dollar_object, header[key], dollar_context)

            elif type(header[key]) == list:
                header[key] = cls.transform(this_dollar_object, header[key], dollar_context)

            elif type(header[key]) == str:
                value = header[key].strip()

                dollar_parsing = False
                dollar_start = 0
                dollar_value = ""
                i = 0
                while i < len(value):
                    char = value[i]
                    if char == "$":
                        if dollar_parsing:
                            raise DollarExecutionException("Already parsing dollar", dollar_context)
                        dollar_parsing = True
                        dollar_start = i
                    elif dollar_parsing:
                        if char == " ":
                            dollar_parsing = False
                            dollar_end = i - 1
                            value = cls._handle_value_string_parse(
                                    value,
                                    header,
                                    dollar_value,
                                    dollar_start,
                                    dollar_end,
                                    dollar_context)
                            i = dollar_end
                            dollar_value = ""
                        else:
                            dollar_value = dollar_value + char
                    i = i + 1
                if dollar_parsing:
                    if dollar_value.startswith("this."):
                        value = cls._handle_value_string_parse(
                                value,
                                header,
                                dollar_value,
                                dollar_start,
                                len(value) - 1,
                                dollar_context)
                    elif dollar_value == "this":
                        value = this_dollar_object
                    elif dollar_start == 0:
                        try:
                            value = DollarObjectIdMap.get(dollar_value)
                        except DollarException as e:
                            raise DollarExecutionException(
                                    e.get_message(),
                                    dollar_context) from e

                header[key] = value

            else:
                raise DollarExecutionException(
                        "Header item of type {} not supported".format(type(header[key])),
                        dollar_context)

        return header

    @classmethod
    def _handle_value_string_parse(cls, value, header, dollar_value, start, end, dollar_context):
        if not dollar_value.startswith("this."):
            raise DollarExecutionException("Dollar reference can only be used locally", dollar_context)
        dollar_key = dollar_value.removeprefix("this.")
        return value[0:start] + header[dollar_key] + value[end + 1:]



from dollar.dollarexception import DollarExecutionException
from dollar.dollarcontext import DollarContext
from dollar.dollarcontext import DollarContextSpan
from dollar.format.raw.rawdollarformat import RawDollarFormatText
from dollar.format.raw.rawdollarformat import RawDollarFormatReference
from dollar.format.raw.rawdollarformat import RawDollarFormatFunction
from dollar.format.raw.rawdollarformat import RawDollarFormatBlock
from dollar.format.raw.rawdollarformat import RawDollarFormatUnion


class RawDollarParser:

    @classmethod
    def parse(cls, raw_content: str, dollar_context_start: DollarContext):
        dollar_count = 0
        dollar_open = 0
        quotes = False
        parenteses = 0
        result = []
        text = ""
        last_char = ""
        file = dollar_context_start.get_file()
        line = dollar_context_start.get_line()
        col = dollar_context_start.get_col()
        last_col = dollar_context_start.get_col()
        dollar_context_start = None
        for i in range(0, len(raw_content)):
            col = col + 1
            if dollar_context_start is None:
                dollar_context_start = DollarContext(file, line, col)
            char = raw_content[i]
            if char == "\n":
                line = line + 1
                last_col = col - 1
                col = 0
            escape = False
            if last_char == "\\":
                escape = True
                if char == "\\":
                    text = text + "\\"
                    last_char = char
                    continue
            if char == "$" and not escape and not quotes:
                if dollar_open == 3 and raw_content[i+1] != "$" and dollar_count != 2:
                    text = text + "$"
                elif dollar_open not in (0, 3):
                    text = text + "$"
                    last_char = char
                    continue
                dollar_count = dollar_count + 1
                if dollar_open == 0 and text != "":
                    dollar_context = DollarContextSpan(
                            dollar_context_start,
                            DollarContext(file, line, col))
                    result.append(cls._create_result_text(text, dollar_context))
                    text = ""
                    dollar_context_start = None
                if dollar_open == dollar_count:
                    if dollar_open == 3:
                        dollar_context = DollarContextSpan(
                                dollar_context_start,
                                DollarContext(file, line, col))
                        result.append(cls._create_result_block(text, dollar_context))
                        text = ""
                        dollar_context_start = None
                    dollar_open = 0
                    dollar_count = 0
            elif char != "\\":
                if dollar_count != 0 and dollar_open == 0:
                    dollar_open = dollar_count
                dollar_count = 0
                if dollar_open in (0, 3):
                    text = text + char
                else:
                    if char == "(":
                        parenteses = parenteses + 1
                    elif char == ")":
                        parenteses = parenteses - 1
                    elif char == "\"" and not escape:
                        quotes = not quotes
                    if char in (" ", "\n") and not escape and not quotes and not parenteses > 0:
                        temp_line = line
                        temp_col = col
                        if col == 0:
                            temp_line = temp_line - 1
                            temp_col = last_col
                        dollar_context = DollarContextSpan(
                                dollar_context_start,
                                DollarContext(file, temp_line, temp_col))
                        if dollar_open == 1:
                            dollar_text = text
                            continue_text = ""
                            if dollar_text[-1] in (".", ",", ":", ";", "!", "?"):
                                continue_text = dollar_text[-1]
                                dollar_text = dollar_text[:-1]
                            result.append(cls._create_result_reference(dollar_text, dollar_context))
                            dollar_open = 0

                            text = continue_text
                            dollar_context_start = DollarContext(file, line, col)
                        elif dollar_open == 2:
                            result.append(cls._create_result_function(text, dollar_context))
                            dollar_open = 0
                            text = ""
                            dollar_context_start = DollarContext(file, line, col)
                    text = text + char
            last_char = char
        if text != "":
            dollar_context = DollarContextSpan(
                    dollar_context_start,
                    DollarContext(file, line, col))
            if dollar_open == 0:
                result.append(cls._create_result_text(text, dollar_context))
            elif dollar_open == 1:
                result.append(cls._create_result_reference(text, dollar_context))
            elif dollar_open == 2:
                result.append(cls._create_result_function(text, dollar_context))
        return result

    @classmethod
    def _create_result_function(cls, text, dollar_context: DollarContext):
        text_split = text.split("(")
        if len(text_split) != 2:
            raise DollarExecutionException("Dollar Function is not formatted properly", dollar_context)
        func_name = text_split[0]
        param_str = text_split[1].split(")")[0]
        col_start_diff = dollar_context.get_col() + len(func_name) + 2
        param_str_split = param_str.split(",")
        params = []
        if len(param_str_split) == 1 and param_str_split[0] == "":
            pass
        else:
            first_param = True
            for param in param_str_split:
                if not first_param:
                    col_start_diff = col_start_diff + 1
                first_param = False
                dollar_context_param_start = DollarContext(
                        dollar_context.get_file(),
                        dollar_context.get_line(),
                        col_start_diff)
                param_content = cls.parse(param, dollar_context_param_start)
                if len(param_content) == 0:
                    dollar_context_param = DollarContextSpan(
                            dollar_context_param_start,
                            DollarContext(
                                    dollar_context.get_file(),
                                    dollar_context.get_line(),
                                    col_start_diff + len(param)))
                    raise DollarExecutionException(
                            "Parameter on {} cant be empty".format(func_name),
                            dollar_context_param)
                if len(param_content) == 1:
                    params.append(param_content[0])
                else:
                    params.append(RawDollarFormatUnion(param_content, dollar_context))
                col_start_diff = col_start_diff + len(param)
        return RawDollarFormatFunction(func_name, params, dollar_context)

    @classmethod
    def _create_result_text(cls, text, dollar_context: DollarContext):
        return RawDollarFormatText(text, dollar_context)

    @classmethod
    def _create_result_reference(cls, text, dollar_context: DollarContext):
        if text[0] == "(" and text[-1] == ")":
            text = text[1:-1]
        return RawDollarFormatReference(text, dollar_context)

    @classmethod
    def _create_result_block(cls, text, dollar_context: DollarContextSpan):
        text_split = text.split("\n")
        plugin_name = text_split[0].strip()
        text_split = text_split[1:]
        contains_tab = True
        for t in text_split:
            if t.strip() != "" and not t.startswith("    "):
                contains_tab = True
        text_split_new = []
        for t in text_split:
            if t.strip() != "":
                if contains_tab:
                    t = t.removeprefix("    ")
                text_split_new.append(t)
        dollar_context_content = DollarContext(
                dollar_context.get_context_start().get_file(),
                dollar_context.get_context_start().get_line() + 1,
                0)
        plugin_content = cls.parse("\n".join(text_split_new), dollar_context_content)
        dollar_context_union = DollarContextSpan(
                DollarContext(
                        dollar_context.get_context_start().get_file(),
                        dollar_context.get_context_start().get_line() + 1,
                        1),
                DollarContext(
                        dollar_context.get_context_end().get_file(),
                        dollar_context.get_context_end().get_line() - 1,
                        len(text_split_new[-1]) + 1)
        )
        return RawDollarFormatBlock(
                plugin_name,
                RawDollarFormatUnion(plugin_content, dollar_context_union),
                dollar_context)



import os
import shutil

from dollar.dollarexception import DollarException


class DollarFileWriter:

    @staticmethod
    def write_files(target_path, outputs):
        if os.path.exists(target_path):
            shutil.rmtree(target_path)
        for output in outputs:
            save_path = os.path.join(target_path, output.get_path())
            save_folder = os.path.split(save_path)[0]
            if not os.path.exists(save_folder):
                os.makedirs(save_folder)
            try:
                with open(save_path, "w") as f:
                    f.write(output.get_content())
            except Exception as e:
                raise DollarException("Could not write file {}".format(save_path)) from e



from dollar.dollarobject import DollarObject
from dollar.format.output.outputformat import OutputFormatText
from dollar.format.output.outputformat import OutputFormatLink
from dollar.format.output.outputformat import OutputFormatDollarObject
from dollar.format.output.outputformatterhandler import OutputFormatterHandler


class OutputFactory:

    @classmethod
    def create_link(cls, this_dollar_object: DollarObject, text, href):
        return OutputFormatterHandler.get_formatter().format_inline(this_dollar_object, [
            OutputFormatLink(href, [
                OutputFormatText(text)
            ])
        ])

    @classmethod
    def create_link_dollar_object(cls, this_dollar_object: DollarObject, to_dollar_object: DollarObject):
        return OutputFormatterHandler.get_formatter().format_inline(this_dollar_object, [
            OutputFormatDollarObject(to_dollar_object)
        ])

    @classmethod
    def create_list(cls, list_structure):
        pass

    @classmethod
    def create_header(cls, text, level):
        pass

    @classmethod
    def create_code_block(cls, code_block_type, code):
        pass

    @classmethod
    def create_plugin_block(cls, plugin_block_type, code):
        return cls.create_code_block(plugin_block_type, code)

    @classmethod
    def create_definition(cls, item, definition):
        pass



from typing import List

from dollar.dollarexception import DollarExecutionException
from dollar.dollarcontext import DollarContext
from dollar.dollarobject import DollarObject
from dollar.helper.validationhelper import ValidationHelper
from dollar.format.input.inputformattype import InputFormatType
from dollar.plugin.dollarplugin import DollarFunctionPlugin
from dollar.plugin.dollarplugin import DollarBlockPlugin


class InputFormat:

    def __init__(self, format_type: InputFormatType, dollar_context: DollarContext):
        self.format_type = format_type
        self.dollar_context = dollar_context

    def get_format_type(self) -> InputFormatType:
        return self.format_type

    def validate(self):
        pass

    def get_dollar_context(self):
        return self.dollar_context

    def __repr__(self):
        return str(self.__dict__)


class InputFormatText(InputFormat):

    def __init__(self, text: str, dollar_context: DollarContext):
        super().__init__(InputFormatType.TEXT, dollar_context)
        self.text = text

    def get_text(self) -> str:
        return self.text

    def validate(self):
        if ValidationHelper.valid_str(self.text):
            raise DollarExecutionException("Text can not be blank", self.dollar_context)


class InputFormatDollarObject(InputFormat):

    def __init__(self, dollar_object: DollarObject, dollar_context: DollarContext):
        super().__init__(InputFormatType.DOLLAR_OBJECT, dollar_context)
        self.dollar_object = dollar_object

    def get_dollar_object(self) -> DollarObject:
        return self.dollar_object

    def validate(self):
        if not isinstance(self.dollar_object, DollarObject):
            raise DollarExecutionException("Dollar object is of wrong type", self.dollar_context)


class InputFormatDollarObjectValue(InputFormatDollarObject):

    def __init__(self, dollar_object: DollarObject, value, dollar_context: DollarContext):
        super().__init__(dollar_object, dollar_context)
        self.format_type = InputFormatType.DOLLAR_OBJECT_VALUE
        self.value = value

    def get_value(self):
        return self.value


class InputFormatFunction(InputFormat):

    def __init__(self,
                 function_plugin: DollarFunctionPlugin,
                 parameters: List[InputFormat],
                 from_dollar_object: DollarObject,
                 dollar_context: DollarContext):
        super().__init__(InputFormatType.FUNCTION, dollar_context)
        self.function_plugin = function_plugin
        self.parameters = parameters
        self.from_dollar_object = from_dollar_object

    def get_function_plugin(self) -> DollarFunctionPlugin:
        return self.function_plugin

    def get_parameters(self) -> List[InputFormat]:
        return self.parameters

    def get_from_dollar_object(self) -> DollarObject:
        return self.from_dollar_object

    def validate(self):
        if not isinstance(self.function_plugin, DollarFunctionPlugin):
            raise DollarExecutionException("Function plugin is of wrong type", self.dollar_context)
        for param in self.parameters:
            param.validate()


class InputFormatBlock(InputFormat):

    def __init__(self, block_plugin: DollarBlockPlugin, content: InputFormat, dollar_context: DollarContext):
        super().__init__(InputFormatType.BLOCK, dollar_context)
        self.block_plugin = block_plugin
        self.content = content

    def get_block_plugin(self) -> DollarBlockPlugin:
        return self.block_plugin

    def get_content(self) -> InputFormat:
        return self.content

    def validate(self):
        if not isinstance(self.block_plugin, DollarBlockPlugin):
            raise DollarExecutionException("Block plugin is of wrong type", self.dollar_context)
        self.content.validate()


class InputFormatUnion(InputFormat):

    def __init__(self, children: List[InputFormat], dollar_context: DollarContext):
        super().__init__(InputFormatType.UNION, dollar_context)
        self.children = children

    def get_children(self) -> List[InputFormat]:
        return self.children

    def validate(self):
        for child in self.children:
            if child.get_format_type() == InputFormatType.UNION:
                raise DollarExecutionException("A union cant contain another union", self.dollar_context)
            child.validate()



from typing import Optional

from dollar.dollarobject import DollarObject


class PluginArg:

    def __init__(self, name: str, arg_type, description: str):
        self.name = name
        self.arg_type = arg_type
        self.description = description

    def get_name(self):
        return self.name

    def get_arg_type(self):
        return self.arg_type

    def get_description(self):
        return self.description


class PluginArgDollarObject(PluginArg):

    def __init__(self, dollar_object_type: Optional[str], description: str):
        super().__init__("dollar_object", DollarObject, description)
        self.dollar_object_type = dollar_object_type

    def get_dollar_object_type(self) -> Optional[str]:
        return self.dollar_object_type



from dollar.confighandler import ConfigHandler
from dollar.confighandler import ConfigType
from dollar.dollarcontext import DollarContext
from dollar.dollarobjectidmap import DollarObjectIdMap
from dollar.dollarobjectimpl import DollarObjectImpl
from dollar.dollarexception import DollarExecutionException
from dollar.dollarexception import DollarException
from dollar.file.dollarfile import DollarFile
from dollar.file.dollarfilereader import DollarFileReader
from dollar.file.dollarfilewriter import DollarFileWriter
from dollar.file.filecopier import FileCopier
from dollar.format.header.headerparser import HeaderParser
from dollar.format.raw.rawdollarparser import RawDollarParser
from dollar.format.transformer.headertransformer import HeaderTransformer
from dollar.format.transformer.inputtostrtransformer import InputToStrTransformer
from dollar.format.transformer.rawtoinputtransformer import RawToInputTransformer
from dollar.plugin.builtinpluginloader import BuiltinPluginLoader
from dollar.plugin.pluginhandler import PluginHandler
from dollar.plugin.pluginmap import PluginMap


class DollarBuilder:

    @staticmethod
    def build():
        ConfigHandler.load_config_default()
        conf_docs_path = ConfigHandler.get(ConfigType.DOCS_PATH)
        conf_target_path = ConfigHandler.get(ConfigType.TARGET_PATH)
        conf_plugin_path = ConfigHandler.get(ConfigType.PLUGIN_PATH)
        conf_file_passthrough = ConfigHandler.get_str_list_opt(ConfigType.FILE_PASSTHROUGH)
        BuiltinPluginLoader.load()
        PluginHandler.import_plugins(conf_plugin_path)

        mdd_files = DollarFileReader.read_mdd_files(conf_docs_path)

        dollar_object_list = []

        for mdd_file in mdd_files:
            path = mdd_file.get_path()
            content = mdd_file.get_content()

            dollar_object = DollarObjectImpl(
                    content,
                    path,
                    path[:-1],
                    "/" + path[:-1].replace("\\", "/"))

            dollar_context = DollarContext(path, 0, 0)

            header_parser_result = HeaderParser.parse(content, dollar_context)
            dollar_object.set_unparsed_header(header_parser_result.header_object)
            dollar_object.set_header_end(header_parser_result.header_end)

            try:
                DollarObjectIdMap.add(dollar_object)
            except DollarException as e:
                raise DollarExecutionException(
                        e.get_message(),
                        dollar_context) from e
            dollar_object_list.append(dollar_object)

        for dollar_object in dollar_object_list:

            dollar_context = DollarContext(dollar_object.get_path(), 0, 0)

            dollar_object.set_header(
                    HeaderTransformer.transform(
                            dollar_object,
                            dollar_object.get_unparsed_header(),
                            dollar_context))
            try:
                primary_plugin = PluginMap.get_extension(dollar_object.get_type())
            except DollarException as e:
                raise DollarExecutionException(
                        e.get_message(),
                        dollar_context) from e
            validate_plugin = primary_plugin
            visited = [dollar_object.get_type()]
            while True:
                if validate_plugin.extends() in visited:
                    raise DollarExecutionException(
                            "Extension plugin {} in Dollar Object {} was caught in a circular dependency loop"
                            .format(primary_plugin.get_name(), dollar_object.get_id()),
                            dollar_context)
                validation = validate_plugin.validate_primary(dollar_object)
                if validation is not None:
                    raise DollarExecutionException(
                            "Validation on {} failed: {}"
                            .format(dollar_object.get_id(), validation),
                            dollar_context)
                if validate_plugin.extends() is not None:
                    visited.append(validate_plugin.extends())
                    try:
                        validate_plugin = PluginMap.get_extension(validate_plugin.extends())
                    except DollarException as e:
                        raise DollarExecutionException(
                                e.get_message(),
                                dollar_context) from e
                else:
                    break

            primary_plugin.exec_primary(dollar_object)

            for key in dollar_object.get_header().keys():
                if PluginMap.has_extension_with_secondary_key(key):
                    PluginMap.get_extension_from_secondary_key(key)\
                            .exec_secondary(dollar_object)

        for dollar_object in dollar_object_list:
            dollar_context_start = DollarContext(
                    dollar_object.get_path(),
                    dollar_object.get_header_end() + 2,
                    0)
            dollar_object.set_raw_formats(
                    RawDollarParser.parse(
                            dollar_object.get_content_without_header(),
                            dollar_context_start))

        for dollar_object in dollar_object_list:
            dollar_object.set_input_formats(
                    RawToInputTransformer.transform_list(dollar_object, dollar_object.get_raw_formats()))

        for dollar_object in dollar_object_list:
            dollar_object.set_output(
                    InputToStrTransformer.transform_list(dollar_object, dollar_object.get_input_formats()))

        outputs = []
        for dollar_object in dollar_object_list:
            dollar_file = DollarFile(
                    dollar_object.get_output_path(),
                    dollar_object.get_output())
            outputs.append(dollar_file)
        DollarFileWriter.write_files(conf_target_path, outputs)

        for file_ending in conf_file_passthrough:
            FileCopier.copy_files(
                    conf_docs_path,
                    conf_target_path,
                    file_ending)

    @classmethod
    def clean_data(self):
        PluginHandler.clean()
        DollarObjectIdMap.clean()



from .rawdollarformat import RawDollarFormat
from .rawdollarformat import RawDollarFormatText
from .rawdollarformat import RawDollarFormatReference
from .rawdollarformat import RawDollarFormatFunction
from .rawdollarformat import RawDollarFormatBlock
from .rawdollarformat import RawDollarFormatUnion
from .rawdollarformattype import RawDollarFormatType
from .rawdollarparser import RawDollarParser



from typing import List

from dollar.plugin.dollarplugin import DollarFunctionPlugin
from dollar.dollarobject import DollarObject
from dollar.format.output.outputformat import OutputFormatText
from dollar.format.output.outputformat import OutputFormatUnorderedList
from dollar.format.output.outputformat import OutputFormatListItem
from dollar.format.output.outputformat import OutputFormatDollarObject
from dollar.plugin.pluginarg import PluginArg


class ListFunctionPlugin(DollarFunctionPlugin):

    def get_name(self):
        return "List"

    def get_arg_info(self) -> List[PluginArg]:
        return [
            PluginArg(
                    "input_list",
                    List,
                    "List which will be printed"),
        ]

    def exec_function(self, input_list):
        output_items = []
        for item in input_list:
            if isinstance(item, DollarObject):
                output_items.append(
                        OutputFormatListItem([
                            OutputFormatDollarObject(item)
                        ]))
            elif type(item) == str:
                output_items.append(
                        OutputFormatListItem([
                            OutputFormatText(item)
                        ]))
        return [OutputFormatUnorderedList(output_items)]



from enum import Enum


class OutputFormatType(Enum):
    TEXT = 0
    BOLD = 1
    ITALIC = 2
    CODE = 3
    LINK = 4
    HEADER = 5
    PARAGRAPH = 6
    TABLE = 7
    TABLE_HEADER = 8
    TABLE_ROW = 9
    ORDERED_LIST = 10
    UNORDERED_LIST = 11
    LIST_ITEM = 12
    CODE_BLOCK = 13
    QUOTE_BLOCK = 14
    DEFINITION = 15
    IMAGE = 16
    PLUGIN_BLOCK = 17
    DOLLAR_OBJECT = 18



from .inputformat import InputFormat
from .inputformat import InputFormatText
from .inputformat import InputFormatDollarObject
from .inputformat import InputFormatDollarObjectValue
from .inputformat import InputFormatFunction
from .inputformat import InputFormatBlock
from .inputformat import InputFormatUnion
from .inputformattype import InputFormatType



from typing import List
from typing import Optional

from dollar.dollarobject import DollarObject
from dollar.plugin.plugintype import PluginType
from dollar.plugin.pluginarg import PluginArg


class DollarPlugin:

    def get_type(self) -> PluginType:
        pass

    def get_name(self) -> str:
        pass


class DollarFunctionPlugin(DollarPlugin):

    def get_type(self) -> PluginType:
        return PluginType.FUNCTION

    def get_arg_info(self) -> List[PluginArg]:
        return []

    def get_name(self) -> str:
        pass

    def exec_function(self, this_dollar_object: DollarObject):
        pass


class DollarBlockPlugin(DollarPlugin):

    def get_type(self) -> PluginType:
        return PluginType.BLOCK

    def get_name(self) -> str:
        pass

    def exec_block(self, something):
        pass


class DollarExtensionPlugin(DollarPlugin):

    def extends(self):
        return None

    def get_type(self) -> PluginType:
        return PluginType.EXTENSION

    def get_name(self) -> str:
        pass

    def validate_primary(self, dollar_object: DollarObject) -> Optional[str]:
        return None

    def get_secondaries(self) -> List[str]:
        return []

    def get_primaries(self) -> List[str]:
        return []

    def exec_primary(self, dollar_object):
        pass

    def exec_secondary(self, dollar_object):
        pass



from typing import List

from dollar.dollarexception import DollarException
from dollar.dollarobject import DollarObject
from dollar.format.output.outputformattype import OutputFormatType


class OutputFormat:

    def __init__(self, format_type: OutputFormatType, children: List['OutputFormat']):
        self.format_type = format_type
        self.children = children
        self.allowed_children = []

    def get_format_type(self):
        return self.format_type

    def get_children(self):
        return self.children

    def is_block(self):
        return False

    def validate(self):
        for child in self.children:
            if child.get_format_type() not in self.allowed_children:
                raise DollarException(
                    "{} is not allowed in {}"
                    .format(child.get_format_type(), self.get_format_type()))


class OutputFormatBlock(OutputFormat):

    def __init__(self, format_type: OutputFormatType, children: List[OutputFormat]):
        super().__init__(format_type, children)

    def is_block(self):
        return True


class OutputFormatText(OutputFormat):

    def __init__(self, text):
        super().__init__(OutputFormatType.TEXT, [])
        self.text = text

    def get_text(self):
        return self.text

    def validate(self):
        if "\n" in self.text:
            raise DollarException(
                "{} is not allowed to contain newline".format(self.get_format_type()))


class OutputFormatBold(OutputFormat):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.BOLD, children)
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.ITALIC,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
        ]


class OutputFormatItalic(OutputFormat):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.ITALIC, children)
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
        ]


class OutputFormatCode(OutputFormatText):

    def __init__(self, text):
        super().__init__(text)
        self.format_type = OutputFormatType.CODE


class OutputFormatLink(OutputFormat):

    def __init__(self, href: str, children: List[OutputFormat]):
        super().__init__(OutputFormatType.LINK, children)
        self.href = href
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.ITALIC
        ]

    def get_href(self):
        return self.href


class OutputFormatHeader(OutputFormatBlock):

    def __init__(self, level: int, children: List[OutputFormat]):
        super().__init__(OutputFormatType.HEADER, children)
        self.level = level
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.ITALIC,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
            OutputFormatType.CODE,
        ]

    def get_level(self):
        return self.level


class OutputFormatParagraph(OutputFormatBlock):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.PARAGRAPH, children)
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.ITALIC,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
            OutputFormatType.CODE,
        ]


class OutputFormatOrderedList(OutputFormatBlock):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.ORDERED_LIST, children)
        self.allowed_children = [
            OutputFormatType.LIST_ITEM,
            OutputFormatType.ORDERED_LIST,
            OutputFormatType.UNORDERED_LIST,
        ]


class OutputFormatUnorderedList(OutputFormatBlock):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.UNORDERED_LIST, children)
        self.allowed_children = [
            OutputFormatType.LIST_ITEM,
            OutputFormatType.ORDERED_LIST,
            OutputFormatType.UNORDERED_LIST,
        ]


class OutputFormatListItem(OutputFormatBlock):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.LIST_ITEM, children)
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.ITALIC,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
            OutputFormatType.CODE,
        ]


class OutputFormatCodeBlock(OutputFormatBlock):

    def __init__(self, code_type, text):
        super().__init__(OutputFormatType.CODE_BLOCK, [])
        self.code_type = code_type
        self.text = text

    def get_code_type(self):
        return self.code_type

    def get_text(self):
        return self.text


class OutputFormatQuoteBlock(OutputFormatBlock):

    def __init__(self, children: List[OutputFormat]):
        super().__init__(OutputFormatType.QUOTE_BLOCK, children)
        self.allowed_children = [
            OutputFormatType.HEADER,
            OutputFormatType.PARAGRAPH,
            OutputFormatType.ORDERED_LIST,
            OutputFormatType.UNORDERED_LIST,
            OutputFormatType.CODE_BLOCK,
            OutputFormatType.QUOTE_BLOCK,
        ]


class OutputFormatDefinition(OutputFormatBlock):

    def __init__(self, defined_children: List[OutputFormat], definition_children: List[OutputFormat]):
        super().__init__(OutputFormatType.DEFINITION, [])
        self.defined_children = defined_children
        self.definition_children = definition_children
        self.allowed_children = [
            OutputFormatType.TEXT,
            OutputFormatType.BOLD,
            OutputFormatType.ITALIC,
            OutputFormatType.LINK,
            OutputFormatType.DOLLAR_OBJECT,
            OutputFormatType.CODE,
        ]

    def get_defined_children(self):
        return self.defined_children

    def get_definition_children(self):
        return self.definition_children

    def validate(self):
        for child in self.defined_children:
            if child.get_format_type() not in self.allowed_children:
                raise DollarException(
                    "{} is not allowed in {}".format(child.get_format_type(), self.get_format_type()))
        for child in self.definition_children:
            if child.get_format_type() not in self.allowed_children:
                raise DollarException(
                    "{} is not allowed in {}".format(child.get_format_type(), self.get_format_type()))


class OutputFormatPluginBlock(OutputFormatBlock):

    def __init__(self, plugin_name, text):
        super().__init__(OutputFormatType.PLUGIN_BLOCK, [])
        self.plugin_name = plugin_name
        self.text = text

    def get_plugin_name(self):
        return self.plugin_name

    def get_text(self):
        return self.text


class OutputFormatDollarObject(OutputFormat):

    def __init__(self, dollar_object: DollarObject):
        super().__init__(OutputFormatType.DOLLAR_OBJECT, [])
        self.dollar_object = dollar_object

    def get_dollar_object(self) -> DollarObject:
        return self.dollar_object

    def get_href(self, this_dollar_object) -> str:
        fr = this_dollar_object.get_target_path().split("/")
        to = self.dollar_object.get_target_path().split("/")
        length = len(fr)
        if length > len(to):
            length = len(to)
        for i in range(0, length-1):
            if fr[0] == to[0]:
                fr = fr[1:]
                to = to[1:]
        result = "/".join(to)
        if len(fr) > 1:
            result = (len(fr)-1) * "../" + result
        else:
            result = "./" + result
        return result



class ValidationHelper:
    @staticmethod
    def valid_obj(o, o_type):
        if o is None or not isinstance(o, o_type):
            return False
        return True

    @staticmethod
    def valid_str(s):
        if s is None or type(s) is not str or s == "":
            return False
        return True



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="dollardoc",
    version="0.2.0",
    description="Object oriented markdown documentation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="dollardoc",
    author_email="development@dollardoc.dev",
    url="https://github.com/dollardoc/dollardoc",
    packages=[
        "dollar",
        "dollar.builder",
        "dollar.cli",
        "dollar.file",
        "dollar.format",
        "dollar.format.header",
        "dollar.format.input",
        "dollar.format.output",
        "dollar.format.raw",
        "dollar.format.transformer",
        "dollar.helper",
        "dollar.plugin",
        "dollar.plugin.builtin",
    ],
    package_dir={"": "src"},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Topic :: Documentation",
        "Topic :: Education",
        "Topic :: Office/Business",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Topic :: Software Development :: Documentation",
        "Topic :: Software Development :: Version Control",
        "Topic :: Text Editors :: Word Processors",
        "Topic :: Text Editors :: Text Processing",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup",
        "Topic :: Text Processing :: Markup :: Markdown",
        "Topic :: Utilities",
    ],
    keywords=[
        "documentation",
        "object-oriented",
        "objectoriented",
        "object oriented",
        "markdown",
        "processing",
        "organize",
        "version",
    ],
    install_requires=[
        "PyYAML"
    ],
    entry_points={
        "console_scripts": [
            "dollardoc=dollar.cli.cli:cli"
        ]
    },
)


from typing import List

from dollar.plugin.dollarplugin import DollarFunctionPlugin
from dollar.dollarobject import DollarObject
from dollar.format.output.outputformat import OutputFormatUnorderedList
from dollar.format.output.outputformat import OutputFormatListItem
from dollar.format.output.outputformat import OutputFormatDollarObject
from dollar.plugin.pluginarg import PluginArg
from dollar.plugin.pluginarg import PluginArgDollarObject


class ListBackLinksFunctionPlugin(DollarFunctionPlugin):

    def get_name(self):
        return "ListBackLinks"

    def get_arg_info(self) -> List[PluginArg]:
        return [
            PluginArgDollarObject(
                    None,
                    "Dollar object of which to list backlinks"),
        ]

    def exec_function(self, dollar_object: DollarObject):
        output = []
        for ref in dollar_object.get_backlinks():
            output.append(OutputFormatListItem([
                OutputFormatDollarObject(ref)
            ]))
        return [
            OutputFormatUnorderedList(output)
        ]



from typing import List

from dollar.plugin.dollarplugin import DollarFunctionPlugin
from dollar.dollarobject import DollarObject
from dollar.format.output.outputformat import OutputFormatText
from dollar.format.output.outputformat import OutputFormatDefinition
from dollar.format.output.outputformat import OutputFormatDollarObject
from dollar.plugin.pluginarg import PluginArg
from dollar.plugin.pluginarg import PluginArgDollarObject


class LinkFunctionPlugin(DollarFunctionPlugin):

    def get_name(self):
        return "Link"

    def get_arg_info(self) -> List[PluginArg]:
        return [
            PluginArgDollarObject(
                    "page",
                    "Dollar object of which to create explicit link to"),
        ]

    def exec_function(self, dollar_object: DollarObject):
        return [
            OutputFormatDefinition([
                OutputFormatDollarObject(dollar_object)
            ], [
                OutputFormatText(dollar_object.get_header()['description'])
            ])
        ]



from typing import List
from typing import cast

from dollar.dollarexception import DollarExecutionException
from dollar.dollarobject import DollarObject
from dollar.format.output.outputformat import OutputFormat
from dollar.format.output.outputformattype import OutputFormatType
from dollar.format.output.outputformat import OutputFormatText
from dollar.format.output.outputformat import OutputFormatBold
from dollar.format.output.outputformat import OutputFormatItalic
from dollar.format.output.outputformat import OutputFormatCode
from dollar.format.output.outputformat import OutputFormatLink
from dollar.format.output.outputformat import OutputFormatHeader
from dollar.format.output.outputformat import OutputFormatParagraph
from dollar.format.output.outputformat import OutputFormatOrderedList
from dollar.format.output.outputformat import OutputFormatUnorderedList
from dollar.format.output.outputformat import OutputFormatListItem
from dollar.format.output.outputformat import OutputFormatCodeBlock
from dollar.format.output.outputformat import OutputFormatQuoteBlock
from dollar.format.output.outputformat import OutputFormatDefinition
from dollar.format.output.outputformat import OutputFormatPluginBlock
from dollar.format.output.outputformat import OutputFormatDollarObject


class OutputFormatterMarkdown:

    @classmethod
    def format(cls, this_dollar_object: DollarObject, dollar_formats: List[OutputFormat], context=""):
        result = []
        for dollar_format in dollar_formats:
            dollar_format.validate()
            dollar_format_type = dollar_format.get_format_type()
            if not dollar_format.is_block():
                raise DollarExecutionException(
                        "{} is not allowed as a root level block".format(dollar_format.get_format_type()))

            if dollar_format_type == OutputFormatType.HEADER:
                dollar_format = cast(OutputFormatHeader, dollar_format)
                head = ""
                for i in range(0, dollar_format.get_level()):
                    head = head + "#"
                result.append(
                        context
                        + head
                        + " "
                        + cls.format_inline(this_dollar_object, dollar_format.get_children()))

            elif dollar_format_type == OutputFormatType.PARAGRAPH:
                dollar_format = cast(OutputFormatParagraph, dollar_format)
                result.append(context + cls.format_inline(this_dollar_object, dollar_format.get_children()))

            elif dollar_format_type == OutputFormatType.ORDERED_LIST:
                dollar_format = cast(OutputFormatOrderedList, dollar_format)
                result.append(
                        context
                        + cls.format_list(this_dollar_object, dollar_format.get_children(), True, context))

            elif dollar_format_type == OutputFormatType.UNORDERED_LIST:
                dollar_format = cast(OutputFormatUnorderedList, dollar_format)
                result.append(
                        context
                        + cls.format_list(this_dollar_object, dollar_format.get_children(), False, context))

            elif dollar_format_type == OutputFormatType.CODE_BLOCK:
                dollar_format = cast(OutputFormatCodeBlock, dollar_format)
                join_str = "\n" + context
                text = join_str.join(dollar_format.get_text().split("\n"))
                result.append(
                        context
                        + "``` "
                        + dollar_format.get_code_type()
                        + "\n"
                        + context
                        + text
                        + "\n"
                        + context
                        + "```")

            elif dollar_format_type == OutputFormatType.QUOTE_BLOCK:
                dollar_format = cast(OutputFormatQuoteBlock, dollar_format)
                result.append(cls.format(this_dollar_object, dollar_format.get_children(), context + "> "))

            elif dollar_format_type == OutputFormatType.DEFINITION:
                dollar_format = cast(OutputFormatDefinition, dollar_format)
                result.append(
                        context
                        + cls.format_inline(this_dollar_object, dollar_format.get_defined_children())
                        + "\n"
                        + context
                        + ": "
                        + cls.format_inline(this_dollar_object, dollar_format.get_definition_children()))

            elif dollar_format_type == OutputFormatType.PLUGIN_BLOCK:
                dollar_format = cast(OutputFormatPluginBlock, dollar_format)
                join_str = "\n" + context
                text = join_str.join(dollar_format.get_text().split("\n"))
                result.append(
                        context
                        + "``` "
                        + dollar_format.get_plugin_name()
                        + "\n"
                        + context
                        + text
                        + "\n"
                        + context
                        + "```")

            else:
                raise DollarExecutionException(
                        "{} is not allowed as a root level block".format(dollar_format.get_format_type()))

        join_str = "\n" + context + "\n"
        return join_str.join(result)

    @classmethod
    def format_inline(cls, this_dollar_object: DollarObject, dollar_formats: List[OutputFormat]):
        result = ""
        for dollar_format in dollar_formats:
            dollar_format.validate()
            dollar_format_type = dollar_format.get_format_type()
            if dollar_format.is_block():
                raise DollarExecutionException(
                        "{} is not allowed as a nested inline item".format(dollar_format.get_format_type()))

            if dollar_format_type == OutputFormatType.TEXT:
                dollar_format = cast(OutputFormatText, dollar_format)
                result = result + dollar_format.get_text()

            elif dollar_format_type == OutputFormatType.BOLD:
                dollar_format = cast(OutputFormatBold, dollar_format)
                result = result + "**" + cls.format_inline(this_dollar_object, dollar_format.get_children()) + "**"

            elif dollar_format_type == OutputFormatType.ITALIC:
                dollar_format = cast(OutputFormatItalic, dollar_format)
                result = result + "*" + cls.format_inline(this_dollar_object, dollar_format.get_children()) + "*"

            elif dollar_format_type == OutputFormatType.CODE:
                dollar_format = cast(OutputFormatCode, dollar_format)
                result = result + "`" + dollar_format.get_text() + "`"

            elif dollar_format_type == OutputFormatType.LINK:
                dollar_format = cast(OutputFormatLink, dollar_format)
                result = result + "["
                result = result + cls.format_inline(this_dollar_object, dollar_format.get_children())
                result = result + "]("
                result = result + dollar_format.get_href()
                result = result + ")"

            elif dollar_format_type == OutputFormatType.DOLLAR_OBJECT:
                dollar_format = cast(OutputFormatDollarObject, dollar_format)
                dollar_object = dollar_format.get_dollar_object()
                title = dollar_object.get_id()
                if "title" in dollar_object.get_header():
                    title = dollar_object.get_header()["title"]
                result = result + "[" + title + "](" + dollar_format.get_href(this_dollar_object) + ")"

            else:
                raise DollarExecutionException("Format {} is not supported".format(dollar_format_type))

        return result

    @classmethod
    def format_list(cls, this_dollar_object: DollarObject, dollar_formats: List[OutputFormat], ordered, context=""):
        count = 1
        result = ""
        for dollar_format in dollar_formats:
            dollar_format.validate()
            dollar_format_type = dollar_format.get_format_type()
            if dollar_format_type == OutputFormatType.LIST_ITEM:
                dollar_format = cast(OutputFormatListItem, dollar_format)
                if count > 1:
                    result = result + "\n"
                if ordered:
                    result = result + context + str(count) + ". "
                else:
                    result = result + context + "* "
                count = count + 1
                result = result + cls.format_inline(this_dollar_object, dollar_format.get_children())

            elif dollar_format_type == OutputFormatType.ORDERED_LIST:
                dollar_format = cast(OutputFormatOrderedList, dollar_format)
                if count == 1:
                    raise DollarExecutionException("List can not be the first item inside a List")
                result = result + "\n" + cls.format_list(
                        this_dollar_object,
                        dollar_format.get_children(),
                        True,
                        context + "    ")

            elif dollar_format_type == OutputFormatType.UNORDERED_LIST:
                dollar_format = cast(OutputFormatUnorderedList, dollar_format)
                if count == 1:
                    raise DollarExecutionException("List can not be the first item inside a List")
                result = result + "\n" + cls.format_list(
                        this_dollar_object,
                        dollar_format.get_children(),
                        False,
                        context + "    ")

            else:
                raise DollarExecutionException("Format {} is not supported on List".format(dollar_format_type))

        return result



from typing import List
from typing import cast

from dollar.helper.dollarobjecthelper import DollarObjectHelper
from dollar.plugin.pluginarg import PluginArg
from dollar.plugin.pluginarg import PluginArgDollarObject
from dollar.dollarexception import DollarException
from dollar.dollarobject import DollarObject


class PluginArgValidator:

    @staticmethod
    def validate(args: List, args_info: List[PluginArg]):
        if len(args) != len(args_info):
            raise DollarException(
                    "Provided {} argument, {} needed"
                    .format(len(args), len(args_info)))
        for i in range(0, len(args)):
            arg = args[i]
            arg_info = args_info[i]
            if not isinstance(arg, arg_info.get_arg_type()):
                raise DollarException(
                        "Type missmatch of argument #{}, {} is not of type {}"
                        .format(i+1, type(arg), arg_info.get_arg_type()))
            if isinstance(arg_info, PluginArgDollarObject):
                dollar_object = cast(DollarObject, arg)
                arg_info = cast(PluginArgDollarObject, arg_info)
                dollar_object_type = arg_info.get_dollar_object_type()
                if dollar_object_type is not None:
                    if not DollarObjectHelper.is_type(dollar_object, dollar_object_type):
                        raise DollarException(
                                "DollarObject with id {}, was not of expected type {}"
                                .format(dollar_object.get_id(), dollar_object_type))



from dollar.plugin.dollarplugin import DollarExtensionPlugin
from dollar.dollarobject import DollarObject


class PageExtensionPlugin(DollarExtensionPlugin):

    def extends(self):
        return None

    def get_name(self):
        return "page"

    def get_secondaries(self):
        return []

    def get_primaries(self):
        return ["title", "description"]

    def validate_primary(self, dollar_object: DollarObject):
        count = 2
        ret = ""

        if "title" not in dollar_object.get_header():
            count = count - 1
            ret = "Page is missing \"title\" key"
        if "description" not in dollar_object.get_header():
            count = count - 1
            ret = "Page is missing \"description\" key"

        if count == 0:
            return "Page is missing \"title\" and \"description\" keys"
        elif count == 1:
            return ret

        return None

    def exec_primary(self, dollar_object: DollarObject):
        pass

    def exec_secondary(self, dollar_object: DollarObject):
        pass



import os
import shutil


class FileCopier:

    @classmethod
    def copy_files(cls, path_from: str, path_to: str, file_ending: str):
        try:
            os.makedirs(path_to)
        except OSError:
            pass  # Allow path to exist

        dir_list = os.listdir(path_from)
        for item in dir_list:
            item_path_from = os.path.join(path_from, item)
            item_path_to = os.path.join(path_to, item)
            if os.path.isdir(item_path_from):
                cls.copy_files(item_path_from, item_path_to, file_ending)
            elif item.endswith(file_ending):
                shutil.copyfile(item_path_from, item_path_to)



import yaml

from dollar.dollarexception import DollarExecutionException
from dollar.dollarcontext import DollarContext
from dollar.format.header.headerparserresult import HeaderParserResult


class HeaderParser:

    @classmethod
    def parse(cls, content, dollar_context: DollarContext):
        content = content.split("\n")
        result = HeaderParserResult(None, 0)

        started_headers = False
        first_line = True
        first_line_in_header = True
        header = ""
        for i in range(0, len(content)):
            line = content[i]

            if line.rstrip() == "---":
                first_line = False
                if started_headers:
                    started_headers = False
                    result.header_end = i
                    break
                else:
                    started_headers = True

            elif first_line:
                if line.strip() != "":
                    break

            else:
                if first_line_in_header:
                    first_line_in_header = False
                else:
                    header = header + "\n"
                header = header + line

        if started_headers:
            raise DollarExecutionException(
                    "Parsing headers exited with a faulty state",
                    dollar_context)

        result.header_object = yaml.safe_load(header)
        return result



class DollarObject:

    def __init__(self, path, output_path, target_path):
        self.path = path
        self.output_path = output_path
        self.target_path = target_path
        self.header = {}
        self.backrefs = []
        self.backlinks = []

    def get_id(self):
        return self.get_header()["id"]

    def get_type(self):
        return self.get_header()["type"]

    def get_header(self):
        return self.header

    def get_backrefs(self):
        return self.backrefs

    def add_backref(self, backref_dollar_object):
        if backref_dollar_object not in self.backrefs:
            self.backrefs.append(backref_dollar_object)

    def get_backlinks(self):
        return self.backlinks

    def add_backlink(self, backlink_dollar_object):
        if backlink_dollar_object != self:
            if backlink_dollar_object not in self.backlinks:
                self.backlinks.append(backlink_dollar_object)

    def get_path(self):
        return self.path

    def get_output_path(self):
        return self.output_path

    def get_target_path(self):
        return self.target_path

    def __repr__(self):
        return "DollarObject(id: " + self.get_id() + ")"
